@@ -54,13 +54,15 @@ public:
     struct Drag {
         SurfaceInterface *surface = nullptr;
         QMetaObject::Connection destroyConnection;
-        QMetaObject::Connection pointerPosConnection;
+        QMetaObject::Connection posConnection;
         QMetaObject::Connection sourceActionConnection;
         QMetaObject::Connection targetActionConnection;
         quint32 serial = 0;
     };
     Drag drag;
 
+    QPointer<SurfaceInterface> proxyRemoteSurface;
+
 private:
     DataDeviceInterface *q_func() {
         return reinterpret_cast<DataDeviceInterface*>(q);
@@ -99,10 +101,19 @@ void DataDeviceInterface::Private::startDragCallback(wl_client *client, wl_resou
 
 void DataDeviceInterface::Private::startDrag(DataSourceInterface *dataSource, SurfaceInterface *origin, SurfaceInterface *i, quint32 serial)
 {
-    // TODO: allow touch
-    if (seat->hasImplicitPointerGrab(serial) && seat->focusedPointerSurface() != origin) {
-        wl_resource_post_error(resource, 0, "Surface doesn't have pointer grab");
-        return;
+    SurfaceInterface *focusSurface = origin;
+    if (proxyRemoteSurface) {
+        // origin is a proxy surface
+        focusSurface = proxyRemoteSurface.data();
+    }
+    const bool pointerGrab = seat->hasImplicitPointerGrab(serial) && seat->focusedPointerSurface() == focusSurface;
+    if (!pointerGrab) {
+        // Client doesn't have pointer grab.
+        const bool touchGrab = seat->hasImplicitTouchGrab(serial) && seat->focusedTouchSurface() == focusSurface;
+        if (!touchGrab) {
+            // Client neither has pointer nor touch grab. No drag start allowed.
+            return;
+        }
     }
     // TODO: source is allowed to be null, handled client internally!
     Q_Q(DataDeviceInterface);
@@ -159,6 +170,10 @@ DataOfferInterface *DataDeviceInterface::Private::createDataOffer(DataSourceInte
     if (!resource) {
         return nullptr;
     }
+    if (!source) {
+        // a data offer can only exist together with a source
+        return nullptr;
+    }
     Q_Q(DataDeviceInterface);
     DataOfferInterface *offer = new DataOfferInterface(source, q, resource);
     auto c = q->global()->display()->getConnection(wl_resource_get_client(resource));
@@ -201,7 +216,7 @@ SurfaceInterface *DataDeviceInterface::icon() const
 SurfaceInterface *DataDeviceInterface::origin() const
 {
     Q_D();
-    return d->surface;
+    return d->proxyRemoteSurface ? d->proxyRemoteSurface.data() : d->surface;
 }
 
 DataSourceInterface *DataDeviceInterface::selection() const
@@ -244,9 +259,9 @@ void DataDeviceInterface::drop()
         return;
     }
     wl_data_device_send_drop(d->resource);
-    if (d->drag.pointerPosConnection) {
-        disconnect(d->drag.pointerPosConnection);
-        d->drag.pointerPosConnection = QMetaObject::Connection();
+    if (d->drag.posConnection) {
+        disconnect(d->drag.posConnection);
+        d->drag.posConnection = QMetaObject::Connection();
     }
     disconnect(d->drag.destroyConnection);
     d->drag.destroyConnection = QMetaObject::Connection();
@@ -261,9 +276,9 @@ void DataDeviceInterface::updateDragTarget(SurfaceInterface *surface, quint32 se
         if (d->resource && d->drag.surface->resource()) {
             wl_data_device_send_leave(d->resource);
         }
-        if (d->drag.pointerPosConnection) {
-            disconnect(d->drag.pointerPosConnection);
-            d->drag.pointerPosConnection = QMetaObject::Connection();
+        if (d->drag.posConnection) {
+            disconnect(d->drag.posConnection);
+            d->drag.posConnection = QMetaObject::Connection();
         }
         disconnect(d->drag.destroyConnection);
         d->drag.destroyConnection = QMetaObject::Connection();
@@ -284,11 +299,16 @@ void DataDeviceInterface::updateDragTarget(SurfaceInterface *surface, quint32 se
         }
         return;
     }
+    if (d->proxyRemoteSurface && d->proxyRemoteSurface == surface) {
+        // A proxy can not have the remote surface as target.
+        // TODO: do this for all client's surfaces?
+        return;
+    }
     auto *source = d->seat->dragSource()->dragSource();
     DataOfferInterface *offer = d->createDataOffer(source);
     d->drag.surface = surface;
     if (d->seat->isDragPointer()) {
-        d->drag.pointerPosConnection = connect(d->seat, &SeatInterface::pointerPosChanged, this,
+        d->drag.posConnection = connect(d->seat, &SeatInterface::pointerPosChanged, this,
             [this] {
                 Q_D();
                 const QPointF pos = d->seat->dragSurfaceTransformation().map(d->seat->pointerPos());
@@ -297,16 +317,30 @@ void DataDeviceInterface::updateDragTarget(SurfaceInterface *surface, quint32 se
                 client()->flush();
             }
         );
+    } else if (d->seat->isDragTouch()) {
+        d->drag.posConnection = connect(d->seat, &SeatInterface::touchMoved, this,
+            [this](qint32 id, quint32 serial, const QPointF &globalPosition) {
+                Q_D();
+                Q_UNUSED(id);
+                if (serial != d->drag.serial) {
+                    // different touch down has been moved
+                    return;
+                }
+                const QPointF pos = d->seat->dragSurfaceTransformation().map(globalPosition);
+                wl_data_device_send_motion(d->resource, d->seat->timestamp(),
+                                        wl_fixed_from_double(pos.x()), wl_fixed_from_double(pos.y()));
+                client()->flush();
+            }
+        );
     }
-    // TODO: same for touch
     d->drag.destroyConnection = connect(d->drag.surface, &QObject::destroyed, this,
         [this] {
             Q_D();
             if (d->resource) {
                 wl_data_device_send_leave(d->resource);
             }
-            if (d->drag.pointerPosConnection) {
-                disconnect(d->drag.pointerPosConnection);
+            if (d->drag.posConnection) {
+                disconnect(d->drag.posConnection);
             }
             d->drag = Private::Drag();
         }
@@ -349,6 +383,13 @@ quint32 DataDeviceInterface::dragImplicitGrabSerial() const
     return d->drag.serial;
 }
 
+void DataDeviceInterface::updateProxy(SurfaceInterface *remote)
+{
+    Q_D();
+    // TODO: connect destroy signal?
+    d->proxyRemoteSurface = remote;
+}
+
 DataDeviceInterface::Private *DataDeviceInterface::d_func() const
 {
     return reinterpret_cast<DataDeviceInterface::Private*>(d.data());
